<?php

declare(strict_types=1);

namespace Drupal\search_api_solr_dense_vector\Plugin\search_api\processor;

use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\OperationType\Embeddings\EmbeddingsInput;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Plugin\search_api\data_type\value\TextToken;
use Drupal\search_api\Processor\ProcessorPluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Add dense vectors to the index using AI module providers.
 *
 * @SearchApiProcessor(
 *   id = "solr_densevector",
 *   label = @Translation("Dense Vector"),
 *   description = @Translation("Generates dense vector embeddings using AI module providers for semantic search of text fields."),
 *   stages = {
 *     "preprocess_index" = 100,
 *   },
 * )
 */
class DenseVector extends ProcessorPluginBase implements PluginFormInterface {

  /**
   * The AI provider plugin manager.
   */
  protected AiProviderPluginManager $aiProviderManager;

  /**
   * The logger service.
   */
  protected LoggerChannelInterface $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->aiProviderManager = $container->get('ai.provider');
    $instance->logger = $container->get('logger.factory')->get('search_api_solr_dense_vector');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'ai_provider' => '',
      'ai_model_id' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['ai_provider'] = [
      '#type' => 'select',
      '#title' => t('AI Provider'),
      '#description' => t('Select the AI provider to use for generating embeddings. Only providers that support embeddings are shown.'),
      '#options' => $this->getEmbeddingsProviders(),
      '#default_value' => $this->configuration['ai_provider'] ?? NULL,
      '#empty_option' => t('- Select an AI provider -'),
    ];

    $form['ai_model_id'] = [
      '#type' => 'select',
      '#title' => t('AI Model'),
      '#description' => t('Select the AI model to use for generating embeddings. The model size must match the vector dimension or you will get errors. This is not implemented yet and the actual values come from the AI default settings.'),
      '#options' => [],
      '#default_value' => $this->configuration['ai_model_id'] ?? [],
      '#empty_option' => t('- Select an AI model -'),
    ];

    $form['note'] = [
      '#markup' => $this->t('To change the vector dimension size or similarity function during index time, you have to edit the Solr field type definition.'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $this->setConfiguration($values);
  }

  /**
   * {@inheritdoc}
   */
  public function preprocessIndexItems(array $items) {
    $config = $this->configuration;

    if (empty($config['ai_provider'])) {
      return;
    }

    try {
      $provider = $this->aiProviderManager->createInstance($config['ai_provider']);
    }
    catch (\Exception $e) {
      $this->logger->error('Failed to create AI provider instance: @message', ['@message' => $e->getMessage()]);
      return;
    }

    $default_config = \Drupal::configFactory()->get('ai.settings');

    foreach ($items as $item) {
      foreach ($item->getFields() as $field) {
        if ('solr_densevector' === $field->getType()) {
          $vectors = [];

          foreach ($field->getValues() as $value) {
            // Convert the value to a string to allow creating the embeddings.
            if (is_array($value)) {
              $text = '';
              foreach ($value as $v) {
                $text .= $v instanceof TextToken ? $v->getText() : $v;
              }
            }
            else {
              $text = (string) $value;
            }

            // Create the embeddings from the text.
            if (!empty($text)) {
              try {
                $embedding_input = new EmbeddingsInput($text);
                $embedding_output = $provider->embeddings($embedding_input, $default_config->get("default_providers.embeddings.model_id"));
                $vectors[] = $embedding_output->getNormalized();
              }
              catch (\Exception $e) {
                $this->logger->error('Failed to generate embeddings: @message', ['@message' => $e->getMessage()]);
                continue;
              }
            }
          }

          // Set the values of the field to the vectors.
          if (count($vectors)) {
            $field->setValues($vectors);
          }
        }
      }
    }
  }

  /**
   * Returns a list of valid embedding AI Providers.
   *
   * @return array
   *   The list of providers.
   */
  protected function getEmbeddingsProviders() {
    $providers = [];

    try {
      $providers = $this->aiProviderManager->getProvidersForOperationType('embeddings');

      foreach ($providers as $plugin_id => $definition) {
        $providers[$plugin_id] = $definition['label'];
      }
    } catch (\Exception $e) {
      $this->logger->error('Failed to load AI providers: @message', ['@message' => $e->getMessage()]);
    }

    return $providers;
  }

}
